'''Functions to calculate suspension thermal noise

'''
from __future__ import division
from numpy import pi, imag, zeros_like, zeros
import numpy as np

from .. import const


def susptherm_stage(f, sus, stage_num, direction):
    """Suspension thermal noise.

    :f: frequency array in Hz
    :sus: gwinc suspension structure

    :returns: displacement noise power spectrum at :f:

    :Temp: must either be set for each stage individually, or globally
    in :sus.Temp:.  The latter will be preferred if specified, so if
    you wish to use per-stage tempurature you must remove :sus.Temp:.

    Assumes suspension transfer functions and V-H coupling have been
    pre-calculated and populated into the relevant `sus` fields.

    """
    # Assign Physical Constants
    kB = const.kB

    # Compute the force on the TM along the beamline
    if direction == 'horiz':
        dxdF = sus.hForce[stage_num]

    elif direction == 'vert':
        # vertical to beamline coupling angle
        theta = sus.VHCoupling.theta

        # convert to beam line motion. theta is squared because
        # we rotate by theta into the suspension basis, and by
        # theta to rotate back to the beam line basis
        dxdF = theta**2 * sus.vForce[stage_num]

    # thermal noise (m^2/Hz) for one suspension
    stage = sus.Stage[::-1][stage_num]
    if 'Temp' in stage:
        T = stage.Temp
    else:
        T = sus.Temp
    w = 2*pi*f

    noise = 4 * kB * T * imag(dxdF) / w

    return noise


def susptherm(f, sus):
    nstage = len(sus.Stage)
    noise = zeros_like(f)
    for stage_num in range(nstage):
        noise += susptherm_stage(f, sus, stage_num, 'horiz')
        noise += susptherm_stage(f, sus, stage_num, 'vert')
    return abs(noise)
