from __future__ import division
from numpy import pi, sqrt, sin, cos, tan, real, imag, zeros
import numpy as np
import logging

from . import const
from .struct import Struct


# supported fiber geometries
FIBER_TYPES = [
    'Round',
    'Ribbon',
    'Tapered',
]


def generate_symbolic_tfs(stages=4):
    import sympy as sp

    # construct quad pendulum equation of motion matrix
    ksyms = sp.numbered_symbols('k')
    msyms = sp.numbered_symbols('m')
    w = sp.symbols('w')
    k = [next(ksyms) for n in range(stages)]
    m = [next(msyms) for n in range(stages)]
    A = sp.zeros(stages)
    for n in range(stages-1):
        # mass and restoring forces (diagonal elements)
        A[n, n] = k[n] + k[n+1] - m[n] * w**2
        # couplings to stages above and below
        A[n, n+1] = -k[n+1]
        A[n+1, n] = -k[n+1]
    # mass and restoring force of bottom stage
    A[-1, -1] = k[-1] - m[-1] * w**2

    # want TM equations of motion, so index 4
    b = sp.zeros(stages, 1)
    b[-1] = 1

    # solve linear system
    xsyms = sp.numbered_symbols('x')
    x = [next(xsyms) for n in range(stages)]
    ans = sp.linsolve((A, b), x)
    return ans


def tst_force_to_tst_displ(k, m, f):
    """transfer function for quad pendulum
    
    """
    k0, k1, k2, k3 = k
    m0, m1, m2, m3 = m
    w = 2*pi*f
    X3 = (k2**2*(k0 + k1 - m0*w**2) + (k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2))*(k2 + k3 - m2*w**2))/(-k3**2*(k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2)) + (k3 - m3*w**2)*(k2**2*(k0 + k1 - m0*w**2) - (-k1**2 + (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2))*(k2 + k3 - m2*w**2)))
    return X3


def top_displ_to_tst_displ(k, m, f):
    """transfer function for quad pendulum
    
    """
    k0, k1, k2, k3 = k
    m0, m1, m2, m3 = m
    w = 2*pi*f
    X0 = k1*k2*k3/(k3**2*(k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2)) - (k3 - m3*w**2)*(k2**2*(k0 + k1 - m0*w**2) + (k1**2 - (k0 + k1 - m0*w**2)*(k1 + k2 - m1*w**2))*(k2 + k3 - m2*w**2)))
    return X0 * k0


def getJointParams(sus, n):
    """extract from ifo struct the wire material, blade material, and
    temperature parameters for the upper and lower joints of stage n

    """
    # Here the suspension stage list is reversed so that the last stage
    # in the list is the test mass
    stages = sus.Stage[::-1]
    last_stage = len(stages) - 1
    stage = stages[n]
    if 'Temp' in stage:
        Temp = stage.Temp
    else:
        Temp = sus.Temp

    TempLower = Temp
    if n == 0:
        # In the top stage, upper and lower joints are assumed to have
        # the same temperature
        TempUpper = Temp
    else:
        # In the lower stages, temperature is allowed to vary between
        # upper and lower joints
        # stage.Temp is the temperature of the lower joint of this stage,
        # and the upper joint of the stage below
        if 'Temp' in stages[n-1]:
            TempUpper = stages[n-1].Temp
        else:
            TempUpper = sus.Temp

    ##############################
    # material parameters

    # support wire
    # upper and lower joint material properties may be different
    # due to temperature gradient
    if 'WireMaterialUpper' in stage and 'WireMaterialLower' in stage:
        wireMatUpper = stage.WireMaterialUpper
        wireMatLower = stage.WireMaterialLower
    elif 'WireMaterial' in stage:
        wireMatUpper = stage.WireMaterial
        wireMatLower = stage.WireMaterial
    elif n == last_stage:
        wireMatUpper = 'Silica'
        wireMatLower = 'Silica'
    else:
        wireMatUpper = 'C70Steel'
        wireMatLower = 'C70Steel'

    WireMaterialUpper = sus[wireMatUpper]
    WireMaterialLower = sus[wireMatLower]
    logging.debug('stage {} wires: {}, {}'.format(n, wireMatUpper, wireMatLower))

    # support blade (upper joint only)
    if 'BladeMaterial' in stage:
        bladeMat = stage.BladeMaterial
    else:
        bladeMat = 'MaragingSteel'

    BladeMaterial = sus[bladeMat]

    logging.debug('stage {} blades: {}'.format(n, bladeMat))

    ULparams = ((0, TempUpper, WireMaterialUpper, BladeMaterial),
                (1, TempLower, WireMaterialLower, None))
    return ULparams


def suspQuad(f, sus):
    """Suspension for quadruple pendulum

    `f` is frequency vector, `sus` is suspension model.  `material`
    specifies material used for test mass suspension stage.  steel
    used for all other stages.  Violin modes are included.

      sus.FiberType should be: 0=round, 1=ribbons.

    hForce, vForce = transfer functions from the force on the TM to TM
    motion these should have the correct losses for the mechanical
    system such that the thermal noise is:

    dxdF = force on TM along beam line to position of TM along beam line
         = hForce + theta^2 * vForce
         = admittance / (i * w)

    where theta = sus.VHCoupling.theta.

    Since this is just suspension thermal noise, the TM internal modes
    and coating properties should not be included.

    hTable, vTable = TFs from support motion to TM motion

    Ah = horizontal equations of motion
    Av = vertical equations of motion

    Adapted from code by Morag Casey (Matlab) and Geppo Cagnoli
    (Maple).  Modification for the different temperatures between the
    stages by K.Arai.

    """
    w = 2 * pi * f
    g = const.g

    # NOTE: For historical reasons, the IFO struct numbers stages from the
    # bottom up (0 is the TM), while this code numbers them from the top down
    # (3 is the TM).  Here the suspension stage list is reversed so that the
    # last stage in the list is the test mass
    stages = sus.Stage[::-1]
    last_stage = len(stages) - 1

    # bottom stage fiber type
    FiberType = sus.FiberType
    assert FiberType in FIBER_TYPES

    ##############################
    # Compute cumulative weight of suspension
    masses = np.array([stage.Mass for stage in stages])
    # weight support by lower stages
    Mgs = g * np.flipud(np.cumsum(np.flipud(masses)))

    ##############################
    # Computation of complex spring constants
    khr = np.zeros([len(stages), len(w)])
    khi = np.zeros([len(stages), 2, len(w)])
    kh  = np.zeros([len(stages), len(w)], dtype=np.complex_)
    kvr = np.zeros([len(stages), len(w)])
    kvi = np.zeros([len(stages), 2, len(w)])
    kv  = np.zeros([len(stages), len(w)], dtype=np.complex_)

    Temps = np.zeros(len(stages))

    for n, stage in enumerate(stages):

        ##############################
        # main suspension parameters
        mass = stage.Mass
        Mg = Mgs[n]
        length = stage.Length

        # Correction for the pendulum restoring force
        kh0 = Mg / length   # N/m, horiz. spring constant, stage n

        kv0_blade = stage.K    # N/m, vert. spring constant (from blade)

        if not np.isnan(stage.WireRadius):
            r_w = stage.WireRadius
        else:
            if 'FiberRadius' in stage:
                r_w = stage.FiberRadius
            elif FiberType == 'Ribbon':
                r_w = sus.Ribbon.Width
            else:
                r_w = sus.Fiber.Radius

        # blade thickness
        t_b = stage.Blade
        # number of support wires
        N_w = stage.NWires

        # wire geometry
        tension = Mg / N_w           # Tension
        xsect = pi * r_w**2          # cross-sectional area
        # cross-sectional moment of inertia
        # Ref: Gretarsson et al, Phys. Lett. A 270 (2000) 108-114 Eq.(2)
        xII = r_w**4 * pi / 4        # x-sectional moment of inertia
        # Surface / Volume
        # Surface = 2*pi*r*h
        # Volume  = pi*r^2*h
        # S/V = 2/r
        mu_v = 2 / r_w               # surface to volume ratio, vertical (wire)
        # Ref: Gretarsson and Harry, Rev. Sci. Inst. 70 (1999) 4081-4087
        # Appendix (A8)
        # There is a geometrical factor \mu to calculate the ratio of the
        # energies between the ones stored in surface and the bulk.
        # For a transversely oscillating fiber of circular cross section, this
        # factor is \mu = 2 as shown in (A11)
        mu_h = mu_v * 2              # surface to volume ratio, horizontal
        # deal with ribbon geometry for last stage
        if n == last_stage and FiberType == 'Ribbon':
            W = sus.Ribbon.Width
            t = sus.Ribbon.Thickness
            xsect = W * t                   # cross-sectional area
            xII = (W * t**3)/12             # x-sectional moment of inertia
            # Surface / Volume
            # Surface = 2*(W+t)*h
            # Volume  = W*t*h
            # S/V = 2*(W+t)/(W*t)
            mu_v = 2 * (W + t) / (W * t)
            # Ref: Gretarsson et al, Phys. Lett. A 270 (2000) 108-114
            # There is a geometrical factor \mu (same as above).
            # This factor is shown in Eq.(6)
            mu_h = mu_v * (3 * N_w * W + t) / (N_w * W + t)
        # deal with tapered geometry for last stage
        if n == last_stage and FiberType == 'Tapered':
            r_end = sus.Fiber.EndRadius
            # recompute these for
            xsectEnd = pi * r_end**2   # cross-sectional area (for delta_h)
            xII = pi * r_end**4 / 4    # x-sectional moment of inertia
            mu_h = 4 / r_end           # surface to volume ratio, horizontal


        ##############################
        # Thermoelastic calculations for wires and blades
        # repeated for upper and lower joint of each stage
        jointParams = getJointParams(sus, n)
        for (isLower, Temp, wireMat, bladeMat) in jointParams:
            alpha_w = wireMat.Alpha  # coeff of thermal expansion
            beta_w = wireMat.dlnEdT  # temp dependence of Young's modulus
            rho_w = wireMat.Rho      # mass density
            C_w = wireMat.C          # heat capacity
            K_w = wireMat.K          # thermal conductivity W/(m K)
            Y_w = wireMat.Y          # Young's modulus
            phi_w = wireMat.Phi      # loss angle

            # surface loss dissipation depth
            # About the surface loss depth see Ref:
            # A. M. Gretarsson and G. M. Harry
            # Rev. Sci. Instrum. 70 (1999) P.4081-4087
            if 'Dissdepth' in wireMat:
                ds_w = wireMat.Dissdepth
            else:
                ds_w = 0             # ignore surface effects

            if not isLower:  # blade parameters (there is no lower blade)
                alpha_b = bladeMat.Alpha # coeff of thermal expansion
                beta_b = bladeMat.dlnEdT # temp dependence of Young's modulus
                rho_b = bladeMat.Rho     # mass density
                C_b = bladeMat.C         # heat capacity
                K_b = bladeMat.K         # thermal conductivity W/(m K)
                Y_b = bladeMat.Y         # Young's modulus
                phi_b = bladeMat.Phi     # loss angle

            # horizontal TE time constant, wires
            # The constant 7.37e-2 is 1/(4*q0^2) from eq 12, C. Zener 10.1103/PhysRev.53.90 (1938)
            # tau_h = 7.37e-2 * 4 * (rho_w * C_w * xsect) / (pi * K_w)
            q0 = 1.84
            tau_h = (rho_w * C_w * xsect) / (q0**2 * pi * K_w)

            # deal with ribbon geometry for last stage
            if n == last_stage and FiberType == 'Ribbon':
                tau_h = (rho_w * C_w * t**2) / (K_w * pi**2)

            # delta: TE factor
            # The first term expresses the cancellation of the CTE and dY/dT
            # beta = (dY/dT)/Y
            # Refs:
            # Cagnoli G and Willems P 2002 Phys. Rev. B 65 174111
            # Cumming A V et al 2009 Class. Quantum Grav. 26 215012
            # horizontal delta, wires
            delta_h = (alpha_w - tension * beta_w / (xsect * Y_w))**2 * Y_w * Temp / (rho_w * C_w)

            # deal with tapered geometry for last stage
            if n == last_stage and FiberType == 'Tapered':
                # use this xsect for thermo-elastic noise
                delta_h = (alpha_w - tension * beta_w / (xsectEnd * Y_w))**2 * Y_w * Temp / (rho_w * C_w)

            # bending length, and dilution factors
            if 'Dilution' in stage and not np.isnan(stage.Dilution):
                # if a specific dilution factor is given, use it
                dil = stage.Dilution
            else:
                # if not, calculate
                d_bend = sqrt(Y_w * xII / tension)
                dil = length / d_bend

            # vertical TE time constant, blades
            tau_v = (rho_b * C_b * t_b**2) / (K_b * pi**2)

            # vertical delta, blades
            # Here the TE cancellation is ignored
            delta_v = Y_b * alpha_b**2 * Temp / (rho_b * C_b)

            ##############################
            # Loss Calculations for wires and blades
            # Track losses from the upper and lower wire joints separately
            # The factor of 2 comes from that the loss was calculated
            # for both upper and lower joints
            # Blades are always attached at the top, there is no "lower" blade
            # The final stage is handled specially and the spring constant
            # calculation below is not used

            # horizontal loss factor for round wires
            # phi_w: bulk loss (brownian motion)
            # (mu_h*ds_w): surface loss (brownian motion)
            # nominally this term becomes zero for steel wires as ds_w is zero
            # The last term is for TE loss
            phih = phi_w * (1 + mu_h * ds_w) + delta_h * tau_h * w / (1 + w**2 * tau_h**2)

            # complex spring constant, horizontal
            khr[n, :] = kh0
            khi[n, isLower, :] = kh0 * phih / dil / 2

            if not isLower:
                # vertical loss factor, blades
                # The bulk loss term and the TE loss term
                phiv_blade = phi_b + delta_v * tau_v * w / (1 + w**2 * tau_v**2)

                # complex spring constant, vertical
                kvr[n, :] = kv0_blade
                kvi[n, 0, :] = kv0_blade * phiv_blade

        ##############################
        # last suspension stage
        # Equations from "GG" (maybe?)
        #   Suspensions thermal noise in the LIGO gravitational wave detector
        #   Gabriela Gonzalez, https://doi.org/10.1088/0264-9381/17/21/305
        # See also "GS"
        #   Brownian motion of a mass suspended by an anelastic wire
        #   Gonzalez & Saulson, https://doi.org/10.1121/1.410467
        #
        # Note:
        #  I in GG = xII
        #  rho in GG = rho_w * xsect
        #  delta in GG = d_bend
        #  T in GG = tension
        #
        # This does not deal with temp gradients between upper and lower joint
        ##############################
        if n == last_stage:

            ### Vertical (bounce) ###
            # loss factor, last stage suspension, vertical (no blades)
            phiv = phi_w * (1 + mu_v * ds_w)

            # vertical Young's modulus
            Y_v = Y_w * (1 + 1j * phiv)

            # vertical spring constant, last stage
            k_z = sqrt(rho_w / Y_v) * w

            def vertical_spring4(stage_length, xsect_area):
                """Compute the vertical spring constant for the last stage

                See equations after equation 12 of GG
                """
                return N_w * xsect_area * Y_v * k_z / tan(k_z * stage_length)

            kv4 = vertical_spring4(length, xsect)

            # deal with tapered geometry for last stage
            if FiberType == 'Tapered':
                l_end = sus.Fiber.EndLength
                l_mid = length - 2 * l_end
                kv_mid = vertical_spring4(l_mid, xsect)
                kv_end = vertical_spring4(l_end, xsectEnd)
                kv4 = 1/(2/kv_end + 1/kv_mid)

            if not np.isnan(kv0_blade):  # if blades present, combine two springs
                kv4wire = kv4
                kv4blade = kvr[last_stage, :] + 1j*kvi[last_stage, 0, :]
                kv4 = 1/(1/kv4wire + 1/kv4blade)

            kvr[last_stage, :] = real(kv4)
            kvi[last_stage, 0, :] = imag(kv4)
            kvi[last_stage, 1, :] = 0

            ### Horizontal (pendulum and violins) ###
            # horizontal Young's modulus
            Y_h  = Y_w * (1 + 1j * phih)

            # simplification factors for later calculations
            # Note: a previous version of this suspension thermal noise
            # calculation used a factor "simp3" which was an incorrect
            # approximation for k_e (see GS eq 10).  Here we approximate k_e
            # as 1/delta.  This is equivalent to assuming delta << 1/k,
            # i.e. the bending length of the fiber is much shorter than the
            # 'ideal string' wavelength.
            k4 = sqrt(rho_w * xsect / tension) * w  # k in GG
            d_bend4 = sqrt(Y_h * xII / tension)     # complex d_bend(4)
            dk4 = k4 * d_bend4                      # dk4 << 1 was assumed
            coskl = cos(k4 * length)
            sinkl = sin(k4 * length)

            # numerator, horiz spring constant, last stage
            #   numerator of K_xx in eq 9 of GG
            #     = T k (cos(k L) + k delta sin(k L))
            #   for w -> 0, this reduces to N_w * T * k
            kh4num  = N_w * tension * k4 * (coskl + dk4 * sinkl)

            # denominator, horiz spring constant, last stage
            #   D after equation 8 in GG
            #   D = sin(k L) - 2 k delta cos(k L)
            #   for w -> 0, this reduces to k (L - 2 delta)
            kh4den = sinkl - 2 * dk4 * coskl

            # horizontal spring constant, last stage
            #   K_xx in eq 9 of GG
            khr[last_stage, :] = real(kh4num / kh4den)
            khi[last_stage, 0, :] = imag(kh4num / kh4den) / 2
            khi[last_stage, 1, :] = imag(kh4num / kh4den) / 2

    ###############################################################
    # Equations of motion for the system
    ###############################################################

    # Calculate horizontal/vertical TFs turning on the loss of each
    # temperature region one by one, for the thermal noise calculation
    hForce = np.zeros([len(stages), len(w)], dtype=complex)
    vForce = np.zeros([len(stages), len(w)], dtype=complex)

    for n in range(len(stages)):
        # turn on only the loss of the current temperature region
        # temp region of each stage includes its lower joint
        # and the upper joint of the stage below (if any)
        # region 0 (TOP) also includes its own upper joint
        lossy_region_lower = np.zeros((len(stages),1))
        lossy_region_upper = np.zeros((len(stages),1))
        lossy_region_lower[n] = 1
        if n == 0:
            lossy_region_upper[0] = 1
        if n < last_stage:
            lossy_region_upper[n+1] = 1

        # horizontal
        # only the imaginary part due to the specified region is used.
        k = khr + 1j*(khi[:,0,:]*lossy_region_upper +
                      khi[:,1,:]*lossy_region_lower)
        # calculate TFs
        hForce[n,:] = tst_force_to_tst_displ(k, masses, f)

        # vertical
        # only the imaginary part due to the specified stage is used
        k = kvr + 1j*(kvi[:,0,:]*lossy_region_upper +
                      kvi[:,1,:]*lossy_region_lower)
        # calculate TFs
        vForce[n,:] = tst_force_to_tst_displ(k, masses, f)

    # Calculate horizontal/vertical TFs with all losses on,
    # for the vibration isolation calculation
    k = khr + 1j*(khi[:,0,:] + khi[:,1,:])
    hTable = top_displ_to_tst_displ(k, masses, f)
    k = kvr + 1j*(kvi[:,0,:] + kvi[:,1,:])
    vTable = top_displ_to_tst_displ(k, masses, f)

    return hForce, vForce, hTable, vTable
